from typing import List

from attr import fields

from typedattr.typeutils import AttrsClass


def get_attr_names(cls: AttrsClass) -> List[str]:
    """Get all attribute names of an attrs class."""
    return [att.name for att in fields(cls)]  # noqa



import inspect
from abc import ABCMeta, abstractmethod
from typing import (
    Iterator, Optional, List, ItemsView, KeysView, ValuesView, Any, Dict, Tuple, Union)


class InstanceToClassDelegator(ABCMeta):
    """Metaclass to delegate instance methods to class methods, e.g.:
    __str__(MyClass) will call MyClass._class_str()."""

    def __str__(cls):
        return cls._class_str()

    def __repr__(cls):
        return cls._class_repr()

    def __iter__(cls):
        return cls._class_iter()

    def __len__(cls):
        return cls._class_len()

    def __instancecheck__(cls, instance):
        return cls._class_instancecheck()

    @abstractmethod
    def _class_str(cls):
        pass

    @abstractmethod
    def _class_repr(cls):
        pass

    @abstractmethod
    def _class_iter(cls):
        pass

    @abstractmethod
    def _class_len(cls):
        pass

    @abstractmethod
    def _class_instancecheck(cls):
        pass


class Const(metaclass=InstanceToClassDelegator):
    """
    Class to hold constants. Cannot be instanced. Behaves similar to a dict.
    This is an approach to overcome some weaknesses of enum.Enum
    (e.g. having to use MyEnum.field.value to get the value).

    Examples:
        Simplest example:
        >>> class MyConstants(Const):
        >>>     FIELD = "value"
        >>> print(MyConstants.FIELD)
        value

        Restrict the types of the constants to one or more types, raising errors on mismatch:
        >>> class MyOtherConstants(MyConstants, allowed_types=str):
        >>>     OTHER_FIELD = "other_value"
        >>> class MoreConstants(Const, allowed_types=(str, int)):
        >>>     OTHER_FIELD = "other_value"
        >>>     OTHER_FIELD_2 = 2

    """

    # create the class properties with empty entries for the root parent
    _dict: Dict[str, Dict[str, Any]] = {"ConstantHolder": {}}

    @classmethod
    def _get_dict(cls):
        return cls._dict[cls.__name__]

    @classmethod
    def __dict__(cls) -> Dict[str, Any]:
        return cls._get_dict()

    @classmethod
    def keys(cls) -> KeysView[str]:
        return cls._get_dict().keys()

    @classmethod
    def values(cls) -> ValuesView[Any]:
        return cls._get_dict().values()

    @classmethod
    def items(cls) -> ItemsView[str, Any]:
        return cls._get_dict().items()

    @classmethod
    def get(cls, item: str, default: Any = None) -> Any:
        try:
            val = cls[item]
        except KeyError:
            return default
        return val

    @classmethod
    def __class_getitem__(cls, item: str):
        try:
            val = cls._get_dict()[item]
        except AttributeError as e:
            raise KeyError(f"{item} in {cls}") from e
        return val

    def __getitem__(self, item: str):
        raise RuntimeError("ConstantHolder cannot be instanced")

    @classmethod
    def __len__(cls) -> int:
        return len(cls.keys())

    @classmethod
    def __iter__(cls) -> Iterator[Any]:
        return iter(cls.keys())

    @classmethod
    def __contains__(cls, item: str) -> bool:
        return item in cls.keys()

    @classmethod
    def _class_str(cls) -> str:
        return f"{cls.__name__}({', '.join('='.join((k, repr(v))) for k, v in cls.items())})"

    @classmethod
    def _class_repr(cls) -> str:
        return cls._class_str()

    @classmethod
    def _class_len(cls):
        return len(cls.keys())

    @classmethod
    def _class_iter(cls):
        return iter(cls.keys())

    @classmethod
    def _class_instancecheck(cls, instance):
        return isinstance(instance, str)

    @classmethod
    def __init_subclass__(
            cls, allowed_types: Optional[Union[type, List[type], Tuple[type, ...]]] = None
    ) -> None:
        """
        Setup properties for the public interface when this class is inherited.

        This will be called on nested inheritance as well.

        Args:
            allowed_types: Optionally specify a type or list of types that are allowed for values.
                By default all values are allowed.
        """
        cls._dict[cls.__name__] = {}

        # add parent fields
        for parent_cls in cls.__bases__:
            try:
                cls._dict[cls.__name__].update(cls._dict[parent_cls.__name__])
            except KeyError:
                # ignore errors when parent class is not a ConstantHolder, eg multiple inheritance
                pass

        # loop attributes, check correctness and extend the parent's class properties _keys, _values, _dict.
        for key in cls.__dict__.keys():
            # ignore non-public fields
            if key[0] == "_":
                continue

            # get the value of the constant
            value = getattr(cls, key)

            # ignore classmethods
            if inspect.ismethod(value) and value.__self__ is cls:
                continue

            # # make sure all constants are uppercase
            # assert key == key.upper(), (
            #     f"Constant: {key} in class: {cls.__name__} must be uppercase.")

            # if allowed types is specified, make sure the value types are allowed
            if allowed_types is not None:
                # isinstance errors when fed lists instead of tuple, so convert lists to tuples
                if isinstance(allowed_types, list):
                    allowed_types = tuple(allowed_types)
                if not isinstance(value, allowed_types):
                    raise TypeError(
                        f"Constant: {key} in class: {cls.__name__} must be of type {allowed_types}")

            # update class properties
            cls._dict[cls.__name__][key] = value

    def __init__(self) -> None:
        raise RuntimeError(
            f"Do not instance this class, it's a ConstantHolder: {type(self).__name__}")



"""Useful types"""
from io import IOBase
from pathlib import Path
from typing import Union, Any, Type

from attrs import AttrsInstance
from importlib_resources.abc import Traversable

TensorType = Any
AttrsClass = Type[AttrsInstance]
PathType = Union[str, Path, Traversable]
PathOrIO = Union[PathType, IOBase]

# Subscripted generics cannot be used with class and instance check
# i.e. to use isinstance(obj, the_type) they need to be defined as a tuple
PathTypeCls = (str, Path, Traversable)
PathOrIOCls = (str, Path, Traversable, IOBase)



from ._typedattr import definenumpy, attrs_from_dict
from .attrutils import get_attr_names
from .const import Const

__all__ = ["definenumpy", "attrs_from_dict", "get_attr_names", "Const"]
__version__ = "0.1.20"



"""
Generic utilities for python objects
"""
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping, Iterable
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Type, List

import numpy as np
from attr import has

from .attrutils import get_attr_names


def get_all_base_classes(klass: type) -> List[type]:
    for base in klass.__bases__:
        yield base
        yield from get_all_base_classes(base)


def is_standard_mapping(d: Any) -> bool:
    return isinstance(d, (dict,))


def is_standard_iterable(d: Any) -> bool:
    return isinstance(d, (list, tuple, set, frozenset))


def is_any_mapping(d: Any) -> bool:
    return isinstance(d, Mapping)


def is_any_iterable(d: Any) -> bool:
    return isinstance(d, Iterable)


def is_iterable(d: Any) -> bool:
    return isinstance(d, Iterable) and not isinstance(d, (str, bytes))


class RecursorInterface(metaclass=ABCMeta):
    @abstractmethod
    def is_iterable_fn(self, d: Any) -> bool:
        pass

    @abstractmethod
    def is_mapping_fn(self, d: Any) -> bool:
        pass


class DefaultRecursor(RecursorInterface):
    """Recurses into all iterables and mappings, except str-like objects"""

    def is_iterable_fn(self, d: Any) -> bool:
        return is_iterable(d)

    def is_mapping_fn(self, d: Any) -> bool:
        return is_any_mapping(d)


class StrictRecursor(RecursorInterface):
    """Recurses only into standard iterables and mappings"""

    def is_iterable_fn(self, d: Any) -> bool:
        return is_standard_iterable(d)

    def is_mapping_fn(self, d: Any) -> bool:
        return is_standard_mapping(d)


def modify_nested_object(
        d: Any, modifier_fn: Callable, return_copy: bool = False,
        parser_class: Type[RecursorInterface] = StrictRecursor) -> Any:
    """
    Traverse a python object and apply a modifier function to all leaves.

    Args:
        d: object to modify (e.g. dict)
        modifier_fn: modifier function to apply to all leaves
        return_copy: whether to copy the object before modifying it
        parser_class: which parser to use for iterable and mapping checks

    Examples:
        >>> e_dict = {'a': 1, 'b': {'c': 2, 'd': 3}}
        >>> modify_nested_object(e_dict, lambda x: x + 1)
        {'a': 2, 'b': {'c': 3, 'd': 4}}

    Returns:

    """
    parser = parser_class()

    def _modify_nested_object(d_inner: Any, depth: int = 0) -> Any:
        recursive_fn = partial(_modify_nested_object, depth=depth + 1)
        if return_copy:
            d_inner = deepcopy(d_inner)
        if parser.is_mapping_fn(d_inner):
            for k, v in d_inner.items():
                d_inner[k] = recursive_fn(v)
        elif parser.is_iterable_fn(d_inner):
            for i, v in enumerate(d_inner):
                d_inner[i] = recursive_fn(v)
        else:
            d_inner = modifier_fn(d_inner)
        return d_inner

    return _modify_nested_object(d)


def flatten_dict(d: Any, separator_for_dict='/', separator_for_list='#',
                 parser_class: Type[RecursorInterface] = StrictRecursor):
    """
    Flatten a nested dict by joining nested keys with a separator.

    Args:
        d: dict to flatten
        separator_for_dict: separator to use for nested dict keys
        separator_for_list: separator to use for list indices
        parser_class: which parser to use for iterable and mapping checks

    Examples:
        >>> e_dict = {'a': 1, 'b': {'c': 2, 'd': 3}, 'e': [1, 2, "6"]}
        >>> flatten_dict(e_dict)
        {'a': 1, 'b/c': 2, 'b/d': 3, 'e#0': 1, 'e#1': 2, 'e#2': '6'}

    Returns:
        Flat dict.
    """
    parser = parser_class()

    def _check_key(key_inner):
        assert separator_for_dict not in key_inner and separator_for_list not in key_inner, (
            f"Separators '{separator_for_dict}' and '{separator_for_list}' not allowed in key "
            f"'{key_inner}' when flattening dict.")

    def _flatten_leaf(d_inner, prefix):
        items_inner = []
        if parser.is_mapping_fn(d_inner):
            for k_inner, v_inner in d_inner.items():
                k_inner_str = str(k_inner)
                _check_key(k_inner_str)
                items_inner.extend(_flatten_leaf(
                    v_inner, prefix=f"{prefix}{separator_for_dict}{k_inner_str}"))
        elif parser.is_iterable_fn(d_inner):
            for i, v_inner in enumerate(d_inner):
                items_inner.extend(_flatten_leaf(
                    v_inner, prefix=f"{prefix}{separator_for_list}{i}"))
        else:
            items_inner.append((prefix, d_inner))
        return items_inner

    items = []
    for k, v in d.items():
        items.extend(_flatten_leaf(v, prefix=f"{k}"))
    return dict(items)


def compare_nested_objects(
        d1_outer: Any, d2_outer: Any,
        recursor_class: Type[RecursorInterface] = StrictRecursor,
        rtol=1.e-5, atol=1.e-8, equal_nan=False
) -> List[str]:
    """
    Compare two nested objects (e.g. dicts) and return a list of str describing all differences.

    Args:
        d1_outer: object 1
        d2_outer: object 2
        recursor_class: recursor definition to find nested content
        rtol: see numpy.allclose
        atol: see numpy.allclose
        equal_nan: see numpy.allclose

    Returns:
        List of strings describing the differences between the two objects.
        Empty list means the objects are identical.
    """
    recursor = recursor_class()

    def _compare_nested_objects(d1, d2, depth: int = 0):
        if type(d1) != type(d2):  # pylint: disable=unidiomatic-typecheck
            return [f"{' ' * depth}Type mismatch: {type(d1)} != {type(d2)}"]

        if isinstance(d1, str):
            # leaf - guard against infinite iteration of strings
            return _compare_leaf(d1, d2, depth)

        if recursor.is_mapping_fn(d1):
            all_errors = []
            for k, v in d1.items():
                if k not in d2:
                    all_errors.append(f"{' ' * depth}Key {k} missing in second dict")
                    continue
                all_errors.extend(_compare_nested_objects(v, d2[k], depth + 1))
            for k in d2.keys():
                if k not in d1:
                    all_errors.append(f"{' ' * depth}Key {k} missing in first dict")
            return all_errors

        if recursor.is_iterable_fn(d1):
            all_errors = []
            if len(d1) != len(d2):
                all_errors.append(f"{' ' * depth}Length mismatch: {len(d1)} != {len(d2)}")
            else:
                for i, v in enumerate(d1):
                    all_errors.extend(_compare_nested_objects(v, d2[i], depth + 1))
            return all_errors

        if has(type(d1)):
            # exact definition of the attrs class is not important
            # attribute names and values matching is enough
            atts1 = get_attr_names(type(d1))
            atts2 = get_attr_names(type(d2))
            if atts1 != atts2:
                return [f"{' ' * depth}Attribute names mismatch: {atts1} != {atts2}"]

            all_errors = []
            for att_name in atts1:
                # check if attributes are defined in the same way
                d1_att = getattr(d1, att_name)
                d2_att = getattr(d2, att_name)
                if type(d1_att) != type(d2_att):  # pylint: disable=unidiomatic-typecheck
                    all_errors.append(f"{' ' * depth}Attribute {att_name} type mismatch: "
                                      f"({type(d1_att)}) != ({type(d2_att)})")
                else:
                    all_errors.extend(_compare_nested_objects(d1_att, d2_att, depth + 1))
            return all_errors

        return _compare_leaf(d1, d2, depth, rtol=rtol, atol=atol, equal_nan=equal_nan)

    return _compare_nested_objects(d1_outer, d2_outer)


def _compare_leaf(d1: Any, d2: Any, depth: int, rtol=1.e-5, atol=1.e-8, equal_nan=False
                  ) -> List[str]:
    # at this point the 2 leaves are guaranteed to be the same type
    if isinstance(d1, np.ndarray):
        comp = np.allclose(d1, d2, rtol=rtol, atol=atol, equal_nan=equal_nan)
    else:
        comp = d1 == d2
    if not comp:
        return [f"{' ' * depth}{d1} != {d2}"]
    return []


def check_object_equality(
        d1: Any, d2: Any, recursor_class: Type[RecursorInterface] = StrictRecursor) -> bool:
    """
    Compare two nested objects (e.g. dicts) and return equality as boolean.

    Args:
        d1: object 1
        d2: object 2
        recursor_class: recursor definition to find nested content

    Returns:
        List of strings describing the differences between the two objects.
        Empty list means the objects are identical.
    """
    return len(compare_nested_objects(d1, d2, recursor_class)) == 0


def big_obj_to_short_str(d: Any) -> str:
    """
    Args:
        d: any big object e.g. a dict or a numpy array

    Returns:
        A hopefully short and representative string representation of the object.
    """
    if d is None:
        return str(None)
    class_name = type(d).__name__
    if hasattr(d, "shape"):
        return f"{class_name} shape {d.shape}"
    try:
        return f"{class_name} len {len(d)}"
    except TypeError:
        return f"Object of type {class_name}"



import collections
from functools import partial
from inspect import isclass
from pathlib import Path
from typing import (
    Tuple, Union, Dict, AbstractSet, Iterable, Mapping, Collection, Type, Any, List, Optional)

import attrs
from attr import has, AttrsInstance
from attrs import define, fields_dict, fields, Attribute

try:
    # python>=3.8
    from typing import get_origin, get_args
except ImportError:
    # python<3.8
    from typing_extensions import get_origin, get_args

from typedattr.typeutils import AttrsClass
from typedattr.objutils import check_object_equality, RecursorInterface, StrictRecursor

# default conversions allow to convert instead of raising errors in case of matching types
# e.g. create Path given str, or create float given int
conversion_type = List[Tuple[Tuple[Type, ...], Type]]
default_conversions: conversion_type = [
    ((str,), Path),
    ((int,), float),
]


def print_fn(*args, **kwargs):  # pylint: disable=unused-argument  # noqa
    # print(*args, **kwargs)  # uncomment for debugging
    pass


def definenumpy(maybe_cls: Union[bool, Type] = False, **kwargs):
    """
    @attrs.define decorator with added numpy equality check
    """
    if isclass(maybe_cls):
        # decorator was used without brackets: call it one more time with brackets
        return definenumpy()(maybe_cls)

    frozen = maybe_cls

    def wrap(cls):
        attrs_cls = define(frozen=frozen, eq=False, **kwargs)(cls)
        attrs_cls.__eq__ = check_object_equality
        return attrs_cls

    return wrap


def attrs_from_dict(
        cls: AttrsClass, input_dict_or_attrs_inst: Union[Dict, AttrsInstance],
        strict: bool = True, skip_unknowns: bool = False,
        recursor_class: Type[RecursorInterface] = StrictRecursor,
        conversions: Optional[conversion_type] = None) -> AttrsInstance:
    """
    Create typechecked attrs instance from dictionary or existing attrs instance.

    Args:
        cls: class decorated with @attrs.define or @definetyped
        input_dict_or_attrs_inst: data source, dict or attrs instance.
        strict: whether to do type checking and conversion
        skip_unknowns: whether to skip input fields that are not defined in cls
        recursor_class: recursor definition to find nested content
        conversions: custom conversions, e.g. if target is annotated as Path and value
            is given as str, then convert to Path instead of raising a TypeError.
            default converts int to float and str to Path.
            pass empty list to disable all conversions.

    Returns:
        Instance of cls with values from input_dict_or_attrs_inst
    """
    recursor = recursor_class()
    try:
        return _attrs_from_dict(recursor, cls, input_dict_or_attrs_inst, strict=strict,
                                skip_unknowns=skip_unknowns, conversions=conversions)
    except TypeError as e:
        raise TypeError(f"Error parsing {cls} from {input_dict_or_attrs_inst}, see the original "
                        f"error above.") from e


def _attrs_from_dict(
        recursor: RecursorInterface, cls: AttrsClass, input_dict_or_attrs: Union[Dict, object],
        strict: bool = False, skip_unknowns: bool = False,
        conversions: Optional[conversion_type] = None):
    print_fn(f"Parsing {cls} from {input_dict_or_attrs}")
    input_cls = type(input_dict_or_attrs)
    if has(input_cls):
        # if given an attrs instance, convert it to dict and then
        # parse it for conversions and typechecking
        assert input_cls == cls, (
            f"Mismatch: Trying to parse input of type {input_cls} as type {cls}")
        input_fields_dict = fields_dict(type(input_dict_or_attrs))
        input_dict: Dict[str, Any] = {k: getattr(input_dict_or_attrs, k) for k in input_fields_dict}
        return _attrs_from_dict(
            recursor, input_cls, input_dict, strict=strict, skip_unknowns=skip_unknowns,
            conversions=conversions)

    input_dict: Dict[str, Any] = input_dict_or_attrs
    if input_dict is None:
        if strict:
            raise TypeError(f"Cannot parse None into {cls}")
        return None

    # check whether input and class match
    all_atts: Tuple[Attribute] = fields(cls)
    all_att_names = set(att.name for att in all_atts)
    matching_input, nonmatching_input = {}, {}
    for key, value in input_dict.items():
        if key in all_att_names:
            matching_input[key] = value
        else:
            nonmatching_input[key] = value

    # split into positional and keyword arguments
    in_args, in_kwargs = [], {}
    cls_fields_dict = fields_dict(cls)
    for field_name, field_value in matching_input.items():
        field_att = cls_fields_dict[field_name]
        try:
            is_positional = bool(field_att.default == attrs.NOTHING)
        except ValueError:
            # some field defaults, e.g. numpy arrays do not have comparison defined here
            # but that means that the default is set and the field is not positional
            is_positional = False
        if is_positional:
            in_args.append(field_value)
        else:
            in_kwargs[field_name] = field_value

    # create an attrs instance from the dict
    # the instance will be flat (nested dicts are not resolved yet) and not typechecked.
    attrs_inst = attrs.evolve(cls(*in_args), **in_kwargs)  # noqa

    # typecheck and unfold nested values in the attrs instance
    for att in all_atts:
        name = att.name
        value = getattr(attrs_inst, name)
        typ = att.type
        new_value = _parse_nested(recursor, name, value, typ, strict=strict,
                                  skip_unknowns=skip_unknowns, conversions=conversions)
        setattr(attrs_inst, name, new_value)

    # handle unknown fields
    if len(nonmatching_input) > 0 and not skip_unknowns:
        if strict:
            raise TypeError(f"Keys in input {list(nonmatching_input.keys())} not defined "
                            f"for class {cls} with attributes {sorted(all_att_names)}")
        for key, value in nonmatching_input.items():
            if skip_unknowns:
                break
            # non-strict mode and keep unknowns: try to add it to the class
            try:
                setattr(attrs_inst, key, value)
            except AttributeError as e:
                raise AttributeError(
                    f"Field {key} is missing from configuration {cls}. Either add it to the "
                    f"configuration or decorate with @attrs.define(slots=False) to allow adding "
                    f"unknown fields or set skip_unknowns=True to ignore them.") from e

    print_fn(f"Output of _attrs_from_dict: {attrs_inst}")
    return attrs_inst


def _parse_nested(recursor: RecursorInterface, name, value, typ,
                  strict: bool = True, skip_unknowns: bool = False,
                  conversions: conversion_type = None, depth: int = 0):
    conversions = default_conversions if conversions is None else conversions
    parse_recursive = partial(_parse_nested, recursor, depth=depth + 1, skip_unknowns=skip_unknowns)

    origin = get_origin(typ)
    args = get_args(typ)

    target_type_name = typ.__name__ if hasattr(typ, "__name__") else str(typ)
    value_type = type(value)
    value_type_name = value_type.__name__ if hasattr(value_type, "__name__") else str(value_type)
    err_msg = (f"Could not parse {name}={value} (type {value_type_name}) as type "
               f"{target_type_name} with strict={strict} "
               # f"depth={depth} type annotation origin={origin} args={args}"
               )

    def maybe_raise_typeerorr(full_err_msg):
        if strict:
            raise TypeError(full_err_msg)
        print_fn(f"Caught: {full_err_msg}. Returning as is.")
        return value

    # resolve nested attrclass
    if has(typ):
        return _attrs_from_dict(recursor, typ, value, strict=strict, skip_unknowns=skip_unknowns,
                                conversions=conversions)

    # resolve any
    if typ == Any:
        return value

    # resolve unions (mostly "optional")
    if origin == Union:
        for new_typ in args:
            try:
                # force strict parsing and catch errors to try all different types
                return parse_recursive(name, value, new_typ, strict=True)
            except TypeError:
                pass
        # if we get here, none of the types worked
        return maybe_raise_typeerorr(f"{err_msg}. No type in Union matches.")

    final_list_type = None

    # check tuples
    if origin == tuple:
        if not (len(args) == 2 and args[1] == Ellipsis):
            # resolve fixed tuple
            if not recursor.is_iterable_fn(value):
                return maybe_raise_typeerorr(f"{err_msg}. Expect a sequence, got {type(value)}")
            if len(value) != len(args):
                return maybe_raise_typeerorr(
                    f"{err_msg}. Expect a sequence with length {len(args)}, got length {len(value)}")
            new_tuple = []
            for item, item_type in zip(value, args):
                new_tuple.append(parse_recursive(name, item, item_type, strict=strict))
            return tuple(new_tuple)
        if len(args) == 0:
            # undefined tuple can be anything
            args = [Any, Ellipsis]

    # check dict
    if origin in (dict, collections.defaultdict):
        if not recursor.is_mapping_fn(value):
            return maybe_raise_typeerorr(f"{err_msg}. Expect a mapping, got {type(value)}")
        if len(args) == 0:
            args = [Any, Any]
        dict_keytype, dict_valuetype = args

        if origin == collections.defaultdict:
            # must keep the default factory if exists
            dict_output = collections.defaultdict(getattr(value, "default_factory", None))
        else:
            dict_output = {}

        for key, val in value.items():
            dict_output[parse_recursive(name, key, dict_keytype, strict=strict)] = parse_recursive(
                name, val, dict_valuetype, strict=strict)

        return dict_output

    if origin in [set, frozenset, tuple]:
        # delegate to list parser
        origin, final_list_type = list, origin
        if len(args) == 0:
            args = [Any]
        args = [args[0]]

    # resolve list
    if origin == list:
        if not recursor.is_iterable_fn(value):
            return maybe_raise_typeerorr(f"{err_msg}. Expect iterable, got {type(value)}")
        list_arg_type = args[0]
        list_output = []
        for item in value:
            list_item = parse_recursive(name, item, list_arg_type, strict=strict)
            list_output.append(list_item)
        if not strict:
            # make sure to not change the input value type
            final_list_type = type(value)
        if final_list_type is not None:
            list_output = final_list_type(list_output)
        return list_output

    # try to parse as a regular type (float, etc), this also parses Callable and any other types
    # but only non-nested parsing and no further inspection.
    try:
        if isinstance(value, typ):
            return value
        print_fn(f"Not compatible: {value} as {typ}, {err_msg}")

    except TypeError as e:
        print_fn(f"Caught {e} while trying to parse {value} as {typ}, {err_msg}")

    # show better error message for abstract types
    try:
        is_sc = issubclass(origin, (AbstractSet, Mapping, Collection, Iterable))
    except TypeError:
        is_sc = False
    if is_sc:
        return maybe_raise_typeerorr(f"{err_msg}. Abstract collections are not supported.")

    # add explicit conversions
    for convert_source, convert_target in conversions:
        if not isclass(typ):
            continue
        if issubclass(typ, convert_target) and isinstance(value, convert_source):
            return convert_target(value)

    if strict:
        add_info = ""
        if typ is None:
            add_info = "Untyped fields not allowed in strict mode. "
        raise TypeError(f"{add_info}{err_msg}. Wrong type or type not supported.")
    print_fn(f"{err_msg}. Returning value {value} as-is")
    return value



import traceback


def format_exception(e, with_traceback=False):
    error_str, error_name = str(e), type(e).__name__
    if error_str == "":
        out_str = error_name
    else:
        out_str = f"{error_name}: {error_str}"

    if not with_traceback:
        return out_str

    tb_list = traceback.format_tb(e.__traceback__)
    tb_str = "".join(tb_list)
    return f"{tb_str}{out_str}"



"""
Wrapper for loguru package

https://loguru.readthedocs.io/en/stable/overview.html
https://loguru.readthedocs.io/en/stable/resources/recipes.html#changing-the-level-of-an-existing-handler

Usage:
    from loguru import logger
    from lmbtools.logging.console import configure_logger, SHORTEST_FORMAT

    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info("Hello")
"""
import os
import sys
from copy import deepcopy
from logging import getLevelName
from typing import Union, Optional, List, Any, Dict

from loguru import logger

LevelType = Union[str, int]  # either "DEBUG" or 10
DEFAULT_LOGURU_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
    "<level>{message}</level>")
SHORTER_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
    "<level>{message}</level>")
SHORTEST_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<level>{message}</level>")
BRIGHTBG_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<blue>{name}</blue>:<blue>{function}</blue>:<blue>{line}</blue> "
    "<level>{message}</level>")


def configure_logger(
        level: LevelType = "DEBUG",
        sink=sys.stderr,
        format=SHORTEST_FORMAT,  # noqa
        colorize=True,
        add_sinks: Optional[List[Any]] = None,
        kwargs_handler: Optional[Dict[str, Any]] = None,
        **kwargs: Any
) -> None:
    """
    Configure the loguru logger. For more complex usages, use logger.configure() directly.

    Args:
        level: minimum level to log
        sink: where to write the logs to
        format: message formatting
        colorize: add color codes to the output
        add_sinks: other sinks to add (str will add a file sink)
        kwargs_handler: additional parameters to pass to each handler
        **kwargs: additional parameters to pass to logger.configure()

    References:
        https://loguru.readthedocs.io/en/stable/api/logger.html
        from loguru import _colorizer  # color code reference

    """
    add_sinks = add_sinks if add_sinks is not None else []
    kwargs_handler = kwargs_handler if kwargs_handler is not None else {}
    sinks = [sink] + add_sinks
    handlers = []
    for lsink in sinks:
        params = {"sink": lsink, "format": format, "colorize": colorize,
                  "level": get_level_as_str(level)}
        assert "sink" not in kwargs_handler, f"sink is a reserved key, got {params}"
        params.update(deepcopy(kwargs_handler))
        if isinstance(lsink, str):
            # automatically disable color codes for files
            params["colorize"] = False
        handlers.append(params)

    # fix bug when combining colorama and loguru in conemu console on windows
    if (colorize and os.name == "nt" and os.environ.get("CONEMUANSI") == "ON"
            and sink in (sys.stderr, sys.stdout)):
        print("\r", end="")  # the invisible print here magically fixes the color
    logger.configure(handlers=handlers, **kwargs)


def get_level_as_str(level: LevelType):
    if isinstance(level, str):
        return level.upper()
    if isinstance(level, int):
        return getLevelName(level).upper()
    raise TypeError(f"Level must be str or int, not {type(level)}")


def get_logger_level_from_args(args) -> str:
    if args.verbose:
        return "DEBUG"
    if args.quiet:
        return "WARNING"
    return "INFO"


def get_terminal_size() -> int:
    try:
        n_cols = os.get_terminal_size().columns
    except OSError:
        n_cols = 80
    return n_cols

# # catch logs - code below will raise exceptions for logs. helpful for debugging where
# # random logs are coming from
# import logging
#
#
# class CustomHandler(logging.Handler):
#     def __init__(self):
#         super().__init__()
#
#     def emit(self, record):
#         raise RuntimeError("Caught log message: {}".format(record.getMessage()))
#
#
# err_logger = logging.getLogger()
# err_logger.addHandler(CustomHandler())
